from typing import Any, Type, TypeVar, Optional, Union
from enum import Enum
from pydantic import BaseModel, ConfigDict, computed_field
import numpy as np


class TaggedModel(BaseModel):
    @computed_field(repr=False)
    @property
    def __data_type__(self) -> str:
        return self.__class__.__name__


class LicenseType(Enum):
    _2D = "2d"
    _3D = "3d"
    default = "default"

    def to_dict(self):
        return {"__type__": "LicenseType", "value": self.value}

    def __str__(self):
        return self.value


TensorType = Union[float, list[float], list[list[float]]]
DTensorType = Union[list[list[float]]]


class MaterialProperties(TaggedModel):
    n: Optional[TensorType] = None
    eps: Optional[TensorType] = None
    mu: Optional[TensorType] = None
    d: Optional[DTensorType] = None

    # this is necessary to support np.ndarrays here...
    model_config = ConfigDict(arbitrary_types_allowed=True)


class MaterialSpec(TaggedModel):
    material: Union[str, MaterialProperties]
    theta: Optional[float] = None
    phi: Optional[float] = None
    x: Optional[float] = None
    loss: Optional[float] = None  # dB/m


T = TypeVar("T")

_TYPE_REGISTRY = {
    "MaterialSpec": MaterialSpec,
    "MaterialProperties": MaterialProperties,
}


def register_type(cls: Type[T]) -> Type[T]:
    """Class decorator to register an Exception subclass."""
    name = cls.__name__
    if name in _TYPE_REGISTRY:
        raise ValueError(f"{name} already registered")
    _TYPE_REGISTRY[name] = cls
    return cls


def get_type(name: str) -> Type:
    try:
        return _TYPE_REGISTRY[name]
    except KeyError:
        raise KeyError(f"Unknown exception type: {name}")


def object_from_dict(data: dict[str, Any]) -> Any:
    """
    Reconstructs an exception from its serialized form.
    Expects data["__data_type__"] to be the class name.
    """
    name = data.pop("__data_type__")
    if not name:
        raise KeyError("Missing 'type' in exception data")

    ExcClass = get_type(name)

    return ExcClass(**data)


@register_type
class EModeError(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._custom_fields = []

    def __setattr__(self, name: str, value: Any) -> None:
        if name != "_custom_fields":
            self._custom_fields.append(name)
        return super().__setattr__(name, value)

    def to_dict(self) -> dict:
        d: dict[str, Any] = {
            "__data_type__": self.__class__.__name__,
        }
        for f in self._custom_fields:
            d[f] = getattr(self, f, None)

        return d


@register_type
class ArgumentError(EModeError):
    def __init__(self, msg: str, function: Optional[str], argument: Optional[str]):
        super().__init__(msg)
        self.msg = msg
        self.function = function
        self.argument = argument

    def __str__(self):
        return f'ArgumentError: the argument: ({self.argument}) to function: ({self.function}) had error: "{self.msg}"'


@register_type
class EPHKeyError(EModeError):
    def __init__(self, msg: str, filename: str, key: str):
        super().__init__(msg)
        self.msg = msg
        self.filename = filename
        self.key = key

    def __str__(self):
        return f'EPHKeyError: the key: ({self.key}) doesn\'t exist in the file: ({self.filename}), "{self.msg}"'


@register_type
class FileError(EModeError):
    def __init__(self, msg: str, filename: str):
        super().__init__(msg)
        self.msg = msg
        self.filename = filename

    def __str__(self):
        return f'FileError: the file: "{self.filename}" had error: "{self.msg}"'


@register_type
class LicenseError(EModeError):
    def __init__(self, msg: str, license_type: LicenseType):
        super().__init__(msg, license_type)
        self.msg = msg
        self.license_type = license_type

    def __str__(self):
        return f'LicenseError: you are using license: {self.license_type!s}, error msg: "{self.msg}"'


@register_type
class ShapeError(EModeError):
    def __init__(self, msg: str, shape_name: str):
        super().__init__(msg)
        self.msg = msg
        self.shape_name = shape_name

    def __str__(self):
        return f'ShapeError: error: "{self.msg}" with shape: {self.shape_name}'


@register_type
class NameError(EModeError):
    def __init__(self, msg: str, type: str, name: str):
        super().__init__(msg)
        self.msg = msg
        self.type = type
        self.name = name

    def __str__(self):
        return f'NameError: error: "{self.msg}" for type: {self.type} and name: {self.name}'


@register_type
class NotImplementedError(EModeError):
    def __init__(self, msg: str):
        super().__init__(msg)
        self.msg = msg

    def __str__(self):
        return f"NotImplementedError: {self.msg}"



from .emodeconnection import EMode
from .eph_utils import open_file, get, inspect
from .types import MaterialSpec, MaterialProperties

__all__ = ["EMode", "open_file", "get", "inspect", "MaterialSpec", "MaterialProperties"]



import errno
import platform
from pathlib import Path
import os
import time
from typing import Optional


def config_dir(ensure_exists=True) -> Path:
    h = Path.home()
    s = platform.system()
    base = (
        os.environ.get("XDG_CONFIG_HOME")
        if s != "Windows"
        else os.environ.get("APPDATA")
    ) or (h / "Library" / "Application Support" if s == "Darwin" else h / ".config")
    p = Path(base) / "emode"

    if ensure_exists:
        p.mkdir(parents=True, exist_ok=True)
    return Path(base) / "emode"


def cache_dir(ensure_exists=True) -> Path:
    h = Path.home()
    s = platform.system()
    base = (
        os.environ.get("XDG_CACHE_HOME")
        if s != "Windows"
        else os.environ.get("LOCALAPPDATA")
    ) or (h / "Library" / "Caches" if s == "Darwin" else h / ".cache")
    p = Path(base) / "emode"
    if ensure_exists:
        p.mkdir(parents=True, exist_ok=True)
    return Path(base) / "emode"


class Cache:
    def __init__(self, port_file_label: Optional[str] = None):
        self.port_file_label = port_file_label
        self.cache = cache_dir(True)

    @property
    def port_path(self):
        if self.port_file_label is None:
            raise ValueError("Need to set port_file_label first")
        return self.cache / f"port_{self.port_file_label}.txt"

    @property
    def print_path(self):
        if self.port_file_label is None:
            raise ValueError("Need to set port_file_label first")
        return self.cache / f"emode_output_{self.port_file_label}.txt"

    def write_port_file(self, port):
        with open(self.port_path, "w") as f:
            f.write(str(port))

    def read_port_file(self, port):
        with open(self.port_path, "r") as f:
            return f.read()

    def cleanup(self, tries=20):
        for i in range(tries):
            try:
                self.port_path.unlink(missing_ok=True)
                self.print_path.unlink(missing_ok=True)
                break
            except OSError as e:
                if e.errno == errno.EBUSY:
                    time.sleep(0.5)

    def __del__(self):
        self.cleanup()



import json
import struct
import socket
import base64
import numpy as np
import time
import os
import sys
import platform
from typing import Any, Optional, Union
from logging import getLogger
from .file_utils import Cache
from .types import EModeError, object_from_dict, TaggedModel

logger = getLogger(__name__)


def obj_hook(dct: dict) -> Optional[Union[np.ndarray, dict]]:
    if "__ndarray__" in dct:
        data = base64.b64decode(dct["__ndarray__"])
        if dct["dtype"] == "object":
            return None  # placeholder value
        else:
            return np.frombuffer(data, dct["dtype"]).reshape(dct["shape"])
    return dct


def detect_ide() -> str:
    """Return the name of the IDE or 'command_line'."""
    env = os.environ
    mods = sys.modules

    if "VSCODE_PID" in env:
        return "vscode"
    if env.get("PYCHARM_HOSTED") == "1" or "pydevd" in mods:
        return "pycharm"
    if "SPYDER_ENV" in env or "SPYDER_ARGS" in env or "spyder_kernels" in mods:
        return "spyder"
    if any(name.startswith("idlelib") for name in mods):
        return "idle"
    # Jupyter variants
    try:
        from IPython import get_ipython  # type: ignore

        shell = get_ipython().__class__.__name__
        if shell in ("ZMQInteractiveShell", "IPKernelApp"):
            return "jupyter"
    except Exception:
        pass
    return "command_line"


def detect_platform() -> dict[str, str]:
    """Return OS name, version, and architecture."""
    return {
        "system": platform.system(),
        "release": platform.release(),
        "version": platform.version(),
        "machine": platform.machine(),
    }


def detect_python() -> dict[str, Any]:
    """Return Python version info and IPython flag."""
    info = {
        "python_version": platform.python_version(),
        "python_implementation": platform.python_implementation(),
        "python_version_info": tuple(sys.version_info),
        "in_ipython": False,
    }
    try:
        from IPython import get_ipython  # type: ignore

        info["in_ipython"] = get_ipython() is not None
    except ImportError:
        pass
    return info


def collect_environment() -> dict[str, Any]:
    """Aggregate all environment info."""
    return {
        "ide": detect_ide(),
        "platform": detect_platform(),
        "python": detect_python(),
    }


class EModeClient:
    def __init__(self, cache: Cache, host: str = "127.0.0.1"):
        self.cache = cache
        self.host = host
        self.port: int = 0
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.settimeout(60)

        t = time.perf_counter()
        wait_time = 60
        while wait_time:
            try:
                with open(cache.port_path, "r") as f:
                    self.port = int(f.read())
            except FileNotFoundError:
                pass
            finally:
                cache.port_path.unlink(missing_ok=True)

            if self.port != 0:
                break
            else:
                wait_time -= time.perf_counter() - t
                t = time.perf_counter()
            time.sleep(0.05)

        if wait_time <= 0:
            self.sock.close()
            raise ConnectionError("EMode connection error!")

        time.sleep(0.1)
        self.sock.connect((self.host, self.port))
        self.sock.settimeout(None)
        self.sock.sendall(self.get_connection_string().encode("utf-8"))
        time.sleep(0.1)
        self.connected = True

    def get_connection_string(self):
        try:
            _ = get_ipython()  # type: ignore
            conn_msg = "connected with iPython"
            self.ipython = True
        except NameError:
            conn_msg = "connected with Python"
            self.ipython = False

        if any("VSCODE" in name for name in os.environ):
            ide = "-VSCode"
        elif any("SPYDER" in name for name in os.environ):
            ide = "-Spyder"
        elif any("JPY_" in name for name in os.environ):
            ide = "-Jupyter"
        elif any("PYCHARM" in name for name in os.environ):
            ide = "-PyCharm"
        elif any("HOME" == name for name in os.environ):
            ide = "-IDLE"
        else:
            ide = "-cmd"

        return conn_msg + ide

    def recv(self) -> dict:
        logger.debug("")
        data = self._recv_msg()
        json_data = json.loads(data.decode("utf-8"), object_hook=obj_hook)
        if isinstance(json_data, dict):
            if version := json_data.pop("__version__", None):
                if version == 1:
                    try:
                        o = object_from_dict(json_data)
                    except KeyError:
                        return json_data

                    if isinstance(o, EModeError):
                        raise o
                    else:
                        return o
        return json_data

    def send(self, data):
        logger.debug(f"sending {data=}")
        for kw in data:
            if isinstance(data[kw], np.ndarray):
                data[kw] = np.squeeze(data[kw]).tolist()

            if isinstance(data[kw], list) and len(data[kw]) == 1:
                data[kw] = data[kw][0]

            if isinstance(data[kw], TaggedModel):
                data[kw] = data[kw].model_dump()
        logger.debug(f"serialized {data=}")
        sendjson = json.dumps(data)
        self._send_raw(sendjson)

    def _send_raw(self, data: str):
        logger.debug(f"sending raw data: {data}")
        msg = bytes(data, encoding="utf-8")
        msg = struct.pack(">I", len(msg)) + msg
        self.sock.sendall(msg)

    def _recv_msg(self):
        raw_msglen = self._recv_raw(4)
        msglen = struct.unpack(">I", raw_msglen)[0]
        ret = self._recv_raw(msglen)
        return ret

    def _recv_raw(self, n: int):
        data = bytearray()
        while len(data) < n:
            packet = self.sock.recv(n - len(data))
            if not packet:
                logger.info("expected data but didn't get it")
                raise ConnectionError(
                    f"connection interrupted? expected {n} bytes, but only got {len(data)}"
                )
            data.extend(packet)
        if n <= 200:
            logger.debug(f"exiting: received: {data=}")
        else:
            logger.debug(f"exiting: received {len(data)} bytes")
        return data

    def close(self):
        try:
            self.send({"function": "exit"})
            self.sock.shutdown(socket.SHUT_RDWR)
            self.sock.close()
        except Exception:
            pass
        self.connected = False
